class Powder:
  """
  A class for powders.
  """

  def __init__(self, name: str) -> None:
    """
    Initializes a new instance of the Powder class.
    Args:
        name (str): The name of the powder.
    """
    self.name = name
