class LoaderNoPlateError(Exception):
  pass


class CentrifugeDoorError(Exception):
  pass


class NotAtBucketError(Exception):
  pass


class BucketNoPlateError(Exception):
  pass


class BucketHasPlateError(Exception):
  pass
